Object subclass: #ScratchServer
	instanceVariableNames: 'stage userName serverSocket incomingUDPSocket peerSockets peerNames sensors lastSentValues incomingBroadcasts incomingNarrowcasts outgoingBroadcasts outgoingNarrowcasts in broadcastCache '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scratch-Networking'!
!ScratchServer commentStamp: '<historical>' prior: 0!
This class supports peer-to-peer collaboration between Scratch instances or between Scratch and other applications that want to interact with Scratch.

Scratch 1.3 includes an experimental extension feature that supports interaction between Scratch and other programs.

Three kinds of interaction are supported:

  a. sharing broadcasts (in both directions)
  b. virtual sensors
  c. Scratch's global variables are made visible

Note that peers can read, but not change, each other's global variables.

Non-Scratch applications can use these mechanisms to control Scratch projects by sending broadcasts and virtual sensors available to Scratch. For example, a music sequencer might send Scratch broadcasts to synchronize animations to music. Or, Scratch can be used to control another application or (through some intermediary program) external hardware devices.

Topology

The network topology is a star, with one Scratch acting as the session host. Clients connect to this host. In addition to processing incoming messages from clients, the host forwards a message from one client to all other clients. The advantage of the star topology is that only the host must have a public IP address and be able to accept incoming TCP/IP connections; the clients may be bethind firewalls that do Network Address Translation (NAT) or disallow incoming connections. The disadvantages of the star toplogy are (a) messages must be forwarded by the host, increasing latency and (b) the host cannot leave the session. However, the star topology is eaiser for distributed users to set up and understand.

A common use of this mechanism is to provide additional sensors to Scratch or allow Scratch to control an external program. In that case, latency is typically very low.

Protocol

The experimental extension feature is enabled using the right-button menu on one of the two sensor blocks. When remote sensors are enabled, Scratch listens for connections on port 42001. 

Once a connection is established, messages are sent in both directions over the socket connection.

Each message consists of a four-byte size, most-significant byte first, followed by the message itself:

	<size: 4 bytes><msg: size bytes>

The four-byte size field is not counted as part of the message size. Thus, the empty message is four zero bytes.

The message up to the first whitespace character (any byte <= 32 is considered whitespace) is interpreted to as a case-insensitive command that is used to decide what to do with the rest of the message. Messages may eventually be used to transmit sprites or media, so large amounts of binary binary data in arbitrary formats may follow the initial command string.

The most common messages are human-readable strings made up of the following elements:

  - unquoted words with no embedded whitespace (cat, dog, mouse)
  - quoted strings ("three word string", "embedded ""quotation marks"" are doubled")
  - signed decimal numbers (1, -1, 3.14, -1.2, .1, -.2)
  - booleans (true or false)

Common Commands

The command set will get extended over time, so clients should just skip any commands that they do not understand. Here are the two most useful ones:

	broadcast <string>
	sensor-update [<var-name> <new-value>]

A sensor update command is followed by zero or more (variable name, value) pairs. Variable names are strings. Values are either numbers or strings.

Scratch outputs these commands when broadcasts or global variable changes occur. Scratch also responds to these commands. Broadcast commands sent to Scratch cause a broadcast to occur. Sensor-update commands update the values of virtual sensors available in the sensor block drop-down menu.


-------

Enchanting has added 'Narrowcasts' which are used for interprocess communication (specifically, between the front end and back end) which should not clutter the list of broadcasts.

!


!ScratchServer methodsFor: 'initialization' stamp: 'CWB 5/26/2011 12:59'!
initialize

	userName _ 'anonymous'.
	peerSockets _  OrderedCollection new.
	peerNames _ IdentityDictionary new.
	sensors _ Dictionary new.
	lastSentValues _ Dictionary new.
	incomingBroadcasts _ OrderedCollection new.
	outgoingBroadcasts _ OrderedCollection new.
	broadcastCache _ Set new.
	incomingNarrowcasts _ OrderedCollection new.
	outgoingNarrowcasts _ OrderedCollection new.
! !


!ScratchServer methodsFor: 'accessing' stamp: 'jm 5/14/2009 12:06'!
broadcastsSeen

	^ broadcastCache asArray sort
! !

!ScratchServer methodsFor: 'accessing' stamp: 'jm 2/14/2008 20:43'!
sensorNames

	^ sensors keys asArray sort

! !

!ScratchServer methodsFor: 'accessing' stamp: 'jm 2/14/2008 20:47'!
sensorValueFor: sensorName
	"Answer the value for the given virtual sensor or nil if the given sensor name is not found."

	^ sensors at: sensorName ifAbsent: [nil]

! !

!ScratchServer methodsFor: 'accessing' stamp: 'jm 2/14/2008 18:41'!
stage: aScratchStageMorph

	stage _ aScratchStageMorph.
! !

!ScratchServer methodsFor: 'accessing' stamp: 'jm 2/14/2008 16:08'!
userName

	^ userName
! !

!ScratchServer methodsFor: 'accessing' stamp: 'jm 2/14/2008 16:08'!
userName: aString

	userName _ aString.
! !


!ScratchServer methodsFor: 'server control' stamp: 'jm 5/14/2009 12:15'!
clearCaches

	broadcastCache _ broadcastCache species new.
	sensors _ sensors species new.
	lastSentValues _ lastSentValues species new.
	self resend: 'send-vars' toPeersExcept: nil.
! !

!ScratchServer methodsFor: 'server control' stamp: 'jm 6/22/2008 21:23'!
endScratchSession
	"Close all sockets. If I am hosting, this will end the session for everyone. If I am only a client, then I will leave the session but the session itself can continue to exist."

	self shutdownServer.

! !

!ScratchServer methodsFor: 'server control' stamp: 'jm 6/22/2008 21:15'!
isHosting
	"Answer true if this is the host of a Scratch session."

	^ serverSocket notNil
! !

!ScratchServer methodsFor: 'server control' stamp: 'jm 6/23/2008 08:12'!
joinSessionAt: ipAddressString
	"Add an outgoing connection to the given address. Fail if a connection cannot be made in bounded amount of time. Answer true if the connection was added successfully."

	| addr sock ok |
	addr _ NetNameResolver addressForName: ipAddressString timeout: 5.
	sock _ MessageSocket new.
	ok _ sock connectTo: addr port: ScratchServer portNumber waitSecs: 5.
	ok ifFalse: [sock destroy. ^ false].

	sock sendMessage: 'peer-name ', userName.
	peerSockets add: sock.
	^ true
! !

!ScratchServer methodsFor: 'server control' stamp: 'CWB 12/21/2010 23:47'!
numberOfPeers
	"Answers with the number of peers this client/server has"

	^ peerSockets size.! !

!ScratchServer methodsFor: 'server control' stamp: 'CWB 12/21/2010 23:37'!
peerCountChanged: how
"Called if the number of peers has changed.  'how' should be #dropped or #connected.  Alerts other parts of the system of changes."

	| frame |

	(frame _ stage ownerThatIsA: ScratchFrameMorph) ifNil: [ ^ nil ].

	frame serverPeerCountChanged: how to: peerSockets size.


! !

!ScratchServer methodsFor: 'server control' stamp: 'jm 6/23/2008 09:23'!
sessionInProgress
	"Answer true if this Scratch server has any external connections."

	^ (peerSockets size > 0) | (serverSocket notNil)

! !

!ScratchServer methodsFor: 'server control' stamp: 'jm 6/22/2008 16:11'!
shutdownServer
	"Destroy all open sockets."

	serverSocket ifNotNil: [
		serverSocket destroy.
		serverSocket _ nil].

	incomingUDPSocket ifNotNil: [
		incomingUDPSocket destroy.
		incomingUDPSocket _ nil].

	peerSockets do: [:sock | sock destroy].
	peerSockets _ peerSockets species new.

	peerNames _ peerNames species new.
	sensors _ sensors species new.
	lastSentValues _ lastSentValues species new.
	outgoingBroadcasts _ outgoingBroadcasts species new.
! !

!ScratchServer methodsFor: 'server control' stamp: 'jm 6/22/2008 21:13'!
startHosting
	"Open a socket on my port and start accepting connections."

	Socket initializeNetwork.
	self shutdownServer.

	serverSocket _ Socket new.
	serverSocket listenOn: self class portNumber backlogSize: 20.

	incomingUDPSocket _ Socket newUDP setPort: self class portNumber.
! !

!ScratchServer methodsFor: 'server control' stamp: 'CWB 5/26/2011 14:10'!
stepServer
	"Do one server step."
	"Note: processIncomingCommands clears outgoingBroadcasts, so do sendOutgoingCommands first."

	self sessionInProgress ifFalse: [^ self].
	self acceptNewConnections.
	self sendOutgoingCommands.
	self processIncomingCommands.
	self processIncomingBroadcasts.
	self processIncomingNarrowcasts.
! !


!ScratchServer methodsFor: 'private-server' stamp: 'CWB 12/21/2010 22:58'!
acceptNewConnections
	"Accept new connections, if any. Also process any newly connected outgoing sockets."

	| newSock |
	serverSocket ifNil: [^ self].
	[serverSocket isConnected] whileTrue: [
		newSock _ serverSocket accept.
		peerSockets addLast: (MessageSocket new on: newSock).
		self resendAllVars.
		self peerCountChanged: #connected.
	].
! !

!ScratchServer methodsFor: 'private-server' stamp: 'jm 6/22/2008 21:28'!
processIncomingBroadcasts
	"NOTE: Broadcast events are added to the outgoingBroadcasts queue as a side-effect of re-broadcasting them locally. To avoid having broadcasts bounce from peer to peer forever, the outgoingBroadcasts queue is cleard after incoming commands have been processed."

	stage ifNotNil: [
		incomingBroadcasts do: [:evtName |
			stage broadcastEventNamed: evtName with: 0]].

	incomingBroadcasts _ incomingBroadcasts species new: 100.
	outgoingBroadcasts _ outgoingBroadcasts species new: 100.
! !

!ScratchServer methodsFor: 'private-server' stamp: 'CWB 12/21/2010 23:36'!
processIncomingCommands
	"Process incoming commands from my peers."

	peerSockets copy do: [:sock |
		sock isConnected
			ifTrue: [
				self processCommandFrom: sock.
				sock sendData]
			ifFalse: [
				sock destroy.
				peerSockets remove: sock ifAbsent: [].
				self peerCountChanged: #dropped.
			]
		].

	self processUDPCommands.
! !

!ScratchServer methodsFor: 'private-server' stamp: 'CWB 5/26/2011 14:11'!
processIncomingNarrowcasts
	"NOTE: Broadcast events are added to the outgoingBroadcasts queue as a side-effect of re-broadcasting them locally. To avoid having broadcasts bounce from peer to peer forever, the outgoingBroadcasts queue is cleard after incoming commands have been processed."

	stage ifNotNil: [
		incomingNarrowcasts do: [:evtName |
			stage receiveNarrowcastNamed: evtName]].

	incomingNarrowcasts _ incomingNarrowcasts species new: 100.
	outgoingNarrowcasts _ outgoingNarrowcasts species new: 100.
! !

!ScratchServer methodsFor: 'private-server' stamp: 'jm 6/22/2008 16:54'!
processUDPCommands
	"Process incoming UDP commands."

	| buf n msg |
	incomingUDPSocket ifNil: [^ self].
	buf _ String new: 3000.
	[true] whileTrue: [
		n _ [incomingUDPSocket receiveDataInto: buf] ifError: [0].
		n = 0 ifTrue: [^ self].

		msg _ buf copyFrom: 1 to: n.
		[self dispatch: msg from: nil] ifError: []].
! !

!ScratchServer methodsFor: 'private-server' stamp: 'jm 5/14/2009 20:28'!
queueBroadcast: aString
	"Queue the given broadcast event to be sent to my peers."

	self sessionInProgress ifFalse: [^ self].

	(#(	'Scratch-KeyPressedEvent'
		'Scratch-MouseClickEvent'
		'Scratch-StartClicked'
	) includes: aString) ifTrue: [^ self].

	outgoingBroadcasts add: aString.
! !

!ScratchServer methodsFor: 'private-server' stamp: 'CWB 5/26/2011 13:03'!
queueNarrowcast: aString
	"Queue the given narrowcast event to be sent to the back end."

	self sessionInProgress ifFalse: [^ self].

	outgoingNarrowcasts add: aString.
! !

!ScratchServer methodsFor: 'private-server' stamp: 'CWB 5/26/2011 13:04'!
sendOutgoingCommands
	"Send broadcasts and variable updates to my peers."

	| varUpdateMsg |
	self someVariableHasChanged
		ifTrue: [varUpdateMsg _ self variableUpdateMessage]
		ifFalse: [varUpdateMsg _ nil].

	peerSockets copy do: [:sock |
		sock isConnected
			ifTrue: [
				varUpdateMsg ifNotNil: [sock sendMessage: varUpdateMsg].
				outgoingBroadcasts do: [:evt | sock sendMessage: (self broadcastMessageFor: evt)].
				outgoingNarrowcasts do: [:evt | sock sendMessage: (self narrowcastMessageFor: evt)].
				sock sendData]
			ifFalse: [
				sock destroy.
				peerSockets remove: sock ifAbsent: [].
				self peerCountChanged: #dropped.
			]
		].

	self recordVariableValues.

! !


!ScratchServer methodsFor: 'private-incoming commands' stamp: 'CWB 5/26/2011 13:06'!
dispatch: aString from: requestSock
	"Dispatch an incoming command from a remote Scratch."

	| cmd op |
	aString size = 0 ifTrue: [
		requestSock sendMessage: 'version "ScratchServer 2.0 alpha"'.
		^ self].

	op _ self opcodeFrom: aString.
	op size = 0 ifTrue: [^ self].

	(#('broadcast' 'narrowcast' 'sensor-update' 'peer-name' 'send-vars' 'sprite') includes: op) ifFalse: [^ self].
	'sprite' = op ifTrue: [
		cmd _ self parseMixedMsg: aString.
		self doSprite: cmd from: requestSock.
		^ self resend: aString toPeersExcept: requestSock].

	cmd _ self parse: aString.

	'broadcast' = op ifTrue: [
		self doBroadcast: cmd from: requestSock.
		^ self resend: aString toPeersExcept: requestSock].

	'narrowcast' = op ifTrue: [
		self doNarrowcast: cmd from: requestSock.
		^ self resend: aString toPeersExcept: requestSock].

	'sensor-update' = op ifTrue: [
		self doSensorUpdate: cmd from: requestSock.
		^ self resend: aString toPeersExcept: requestSock].

	'peer-name' = op ifTrue: [^ self doPeerName: cmd from: requestSock].

	'send-vars' = op ifTrue: [
		self doSendVars: cmd from: requestSock.
		^ self resend: aString toPeersExcept: requestSock].

! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'jm 5/14/2009 12:25'!
doBroadcast: cmd from: requestSocket
	"Handle a broadcast command: broadcast <event-name>"

	| evtName |
	cmd size = 2 ifFalse: [^ self].
	((evtName _ cmd at: 2) isKindOf: String) ifFalse: [^ self].

	incomingBroadcasts add: evtName.
	broadcastCache add: evtName.
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'CWB 5/26/2011 14:11'!
doNarrowcast: cmd from: requestSocket
	"Handle a narrowcast command: narrowcast <event-name>"

	| evtName |
	cmd size = 2 ifFalse: [^ self].
	((evtName _ cmd at: 2) isKindOf: String) ifFalse: [^ self].

	incomingNarrowcasts add: evtName.
	"broadcastCache add: evtName."	"No cache for narrowcasts"
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'jm 6/23/2008 09:20'!
doPeerName: cmd from: requestSock
	"Handle a peer-name command: peer-name <peerName>"

	| peerName |
	cmd size = 2 ifFalse: [^ self].
	((peerName _ cmd at: 2) isKindOf: String) ifFalse: [^ self].

	peerNames at: requestSock put: peerName.
	self dispatch: 'send-vars' from: nil.
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'jm 6/23/2008 09:19'!
doSendVars: cmd from: requestSock
	"Handle a variable send request: send-vars"

	"force all my global variables to be re-transmitted"
	lastSentValues _ lastSentValues species new.
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'JM 8/26/2011 13:04'!
doSensorUpdate: cmd from: requestSocket
	"Handle a sensor update command: sensor-update [<sensor-name> <sensor-value>]"

	| i sName sValue |
	i _ 2.
	[i < cmd size] whileTrue: [
		sName _ cmd at: i.
		sValue _ cmd at: i + 1.
		(sName isKindOf: String)
			ifTrue: [
				"Variables updated on the NXT all start with Enc^.  Remote variables from Scratch do not."
				(sName beginsWith: 'Enc^') ifFalse: [
					"Scratch variables act as remote sensors"
					sensors at: sName put: sValue. 
				] ifTrue: [
					stage ifNotNil: [
						| shortName |
						"This is a variable from Enchanting."
						"Strip the header"
						shortName _ sName allButFirst: ('Enc^' size).

						"Update the value of the variable or list"
						(stage listNamed: shortName)
							ifNotNil: [ stage setList: shortName to: sValue ]
							ifNil: [ stage setVar: shortName to: sValue ].

						"Remember the change"
						lastSentValues at: shortName put: sValue.
					].
				].
			].
		i _ i + 2].
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'jens 9/16/2008 09:30'!
doSprite: cmd from: requestSock
	"Handle a sprite command: sprite <senderName> <spriteData>"

	| sf spriteData|
	cmd size = 3 ifFalse: [^ self].
	((spriteData _ cmd at: 3) isKindOf: ByteArray) ifFalse: [^ self].
	sf _ stage ownerThatIsA: ScratchFrameMorph.
	sf ifNil: [^self].
	sf importSpriteOrProject: spriteData
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'jm 8/4/2008 08:06'!
opcodeFrom: aString
	"Answer the command opcode for the given message. The opcode is the unquoted string at the beginning of the message up to the first whitespace character."

	| i |
	i _ 1.
	[i <= aString size and: [(aString at: i) asciiValue > 32]]
		whileTrue: [i _ i + 1].

	^ (aString copyFrom: 1 to: i - 1) asLowercase
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'jm 6/22/2008 16:54'!
processCommandFrom: requestSock
	"Handle a command from the given socket. Do nothing if there are no messages."

	| msg |
	(msg _ requestSock nextMessage) ifNil: [^ self].

	[self dispatch: msg asString from: requestSock]
		ifError: [:err :rcvr | err].
! !

!ScratchServer methodsFor: 'private-incoming commands' stamp: 'jm 6/22/2008 17:05'!
resend: msgString toPeersExcept: requestSocket
	"Handle a broadcast command: broadcast <event-name>"

	peerSockets do: [:sock |
		sock ~= requestSocket ifTrue: [sock sendMessage: msgString]].
! !


!ScratchServer methodsFor: 'private-command parsing' stamp: 'jm 2/14/2008 13:42'!
parse: aString
	"ScratchServer new parse: 'test abc 1 -1 .01 -.01 true false'"

	| out token |
	in _ ReadStream on: aString asString.
	out _ OrderedCollection new.
	[in atEnd] whileFalse: [
		token _ self readNext.
		token ifNotNil: [out addLast: token]].
	^ out asArray
! !

!ScratchServer methodsFor: 'private-command parsing' stamp: 'jens 9/16/2008 09:35'!
parseMixedMsg: aString
	"answer an Array containting both textual and binary information:
	op <senderName> <binaryData>"

	| op senderName i binaryData |
	op _ self opcodeFrom: aString.
	i _ aString findString: String crlf.
	i = 0 ifTrue: [^self error: 'delimiter for binary data not found'].
	senderName _ aString copyFrom: op size + 2 to: i-1.
	binaryData _ aString copyFrom: i + 2  to: aString size.
	^Array with: op with: senderName with: binaryData asByteArray
! !

!ScratchServer methodsFor: 'private-command parsing' stamp: 'jm 2/14/2008 12:41'!
readNext
	"Read the next command line token from 'in' or nil if there are not more tokens. A token is a quoted string, a number, a boolean, or an unquoted string with no embedded whitespace. Strings that start with a digit or minus sign must be quoted to avoid being treated as an integer."

	| ch s |
	"skip white space"
	[in atEnd not and: [in peek asciiValue <= 32]]
		whileTrue: [in next].
	in atEnd ifTrue: [^ nil].

	ch _ in peek.
	ch = $" ifTrue: [^ self readQuotedString].
	(ch = $-) | (ch = $.) | ch isDigit ifTrue: [^ self readNumber].

	"unquoted string or boolean"
	s _ self readString.
	'true' = s ifTrue: [^ true].
	'false' = s ifTrue: [^ false].
	^ s
! !

!ScratchServer methodsFor: 'private-command parsing' stamp: 'jens 7/1/2010 02:13'!
readNumber

	| orig s sign |
	s _ self readString.
	orig _ s.
	sign _ 1.
	s first = $- ifTrue: [s _ s copyFrom: 2 to: s size. sign _ -1].

	s size < 1 ifTrue: [^ orig]. "needed for BYOB, -jens"

	s first = $. ifTrue: [s _ '0', s].  "add leading zero"
	^ s asNumber * sign
! !

!ScratchServer methodsFor: 'private-command parsing' stamp: 'jm 7/31/2008 07:45'!
readQuotedString

	| result ch |
	result _ WriteStream on: String new.
	in next.  "skip opening quote"
	[in atEnd] whileFalse: [
		ch _ in next.
		ch = $" ifTrue: [
			in peek = $"
				ifTrue: [in skip: 1]
				ifFalse: [^ result contents]].
		result nextPut: ch].

	^ result contents
! !

!ScratchServer methodsFor: 'private-command parsing' stamp: 'jm 2/14/2008 12:26'!
readString

	| result ch |
	result _ WriteStream on: String new.
	[in atEnd] whileFalse: [
		ch _ in peek.
		ch asciiValue <= 32 ifTrue: [^ result contents].
		result nextPut: in next].

	^ result contents
! !


!ScratchServer methodsFor: 'private-variables' stamp: 'jm 5/19/2009 10:46'!
broadcastMessageFor: aString
	"Answer a message to broadcast the given string."

	| msg |
	msg _ WriteStream on: String new.
	msg nextPutAll: 'broadcast '.
	self putString: aString on: msg.

	^ msg contents
! !

!ScratchServer methodsFor: 'private-variables' stamp: 'CWB 5/26/2011 13:08'!
narrowcastMessageFor: aString
	"Answer a message to narrowcast the given string."

	| msg |
	msg _ WriteStream on: String new.
	msg nextPutAll: 'narrowcast '.
	self putString: aString on: msg.

	^ msg contents
! !

!ScratchServer methodsFor: 'private-variables' stamp: 'jm 7/31/2008 14:47'!
putString: aString on: aStream
	"Append the given string to the given stream. Double any embedded double-quote characters."

	aStream nextPut: $".
	aString do: [:ch |
		aStream nextPut: ch.
		ch = $" ifTrue: [aStream nextPut: ch]].
	aStream nextPut: $".
! !

!ScratchServer methodsFor: 'private-variables' stamp: 'jm 2/15/2008 08:01'!
recordVariableValues
	"Record the current values of my global variables."

	lastSentValues _ Dictionary new.

	stage ifNotNil: [
		stage varNames do: [:v | lastSentValues at: v put: (stage getVar: v)]].
! !

!ScratchServer methodsFor: 'private-variables' stamp: 'jm 3/26/2009 09:36'!
resendAllVars
	"Force all my global variables to be re-transmitted."

	lastSentValues _ lastSentValues species new.
! !

!ScratchServer methodsFor: 'private-variables' stamp: 'jm 2/15/2008 07:58'!
someVariableHasChanged
	"Answer true if a global variable has changed since I last sent variable updates to my peers."

	| lastVal |
	stage ifNil: [^ false].

	stage varNames do: [:v |
		lastVal _ lastSentValues at: v ifAbsent: [nil].
		lastVal ifNil: [^ true].
		(lastVal = (stage getVar: v)) ifFalse: [^ true]].

	^ false
! !

!ScratchServer methodsFor: 'private-variables' stamp: 'jm 5/14/2009 08:16'!
variableUpdateMessage
	"Answer a variable update message for all variables whose values have not already been sent."

	| msg currentValue lastValue v |
	msg _ WriteStream on: String new.
	msg nextPutAll: 'sensor-update '.
	stage varNames do: [:varName |
		currentValue _ stage getVar: varName.
		lastValue _ lastSentValues at: varName ifAbsent: [nil].
		lastValue = currentValue ifFalse: [
			self putString: varName on: msg.
			msg space.
			v _ stage interpretStringAsNumberIfPossible: currentValue.
			v isNumber
				ifTrue: [msg nextPutAll: v printString]
				ifFalse: [self putString: v asString on: msg].
			msg space]].

	^ msg contents
! !


!ScratchServer methodsFor: 'private-sprites' stamp: 'jens 9/16/2008 09:33'!
sendSprite: aScratchSpriteMorph
	"send the given sprite to all peers"

	| msg |
	msg _ self spriteMessageFor: aScratchSpriteMorph.
	peerSockets copy do: [:sock |
		sock isConnected
			ifTrue: [
				sock sendMessage: msg.
				sock sendData]
			ifFalse: [
				sock destroy.
				peerSockets remove: sock ifAbsent: []]].! !

!ScratchServer methodsFor: 'private-sprites' stamp: 'jens 9/16/2008 09:32'!
spriteMessageFor: aScratchSpriteMorph
	"answer a valid msg for the given sprite: sprite <senderName><crlf><binaryData>"

	|msg|
	msg _ WriteStream on: (ByteArray new: 100000).
	msg 
		nextPutString: 'sprite ';
		nextPutString: userName;
		nextPutString: String crlf;
		nextPutAll: aScratchSpriteMorph serialized.
	^msg contents 
		! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ScratchServer class
	instanceVariableNames: ''!

!ScratchServer class methodsFor: 'utilities' stamp: 'jm 2/16/2008 11:28'!
getIPAddressFromServer
	"Attempt to get my IP address from the AddressServer in bounded time. Returns nil if unsuccessful."
	"[self getIPAddressFromServer] msecs"

	| serverAddr sock msg |
	serverAddr _ #(18 85 18 78) asByteArray.
	sock _ MessageSocket new.
	sock connectTo: serverAddr port: 54330 waitSecs: 4.
	sock isConnected ifFalse: [^ nil].
	sock sendMessage: 'lookup'.
	20 timesRepeat: [
		sock sendData.
		(msg _ sock nextMessage) ifNotNil: [sock destroy. ^ msg].
		Delay waitMSecs: 100].

	"timed out"
	sock destroy.
	^ nil
! !

!ScratchServer class methodsFor: 'utilities' stamp: 'CWB 9/10/2010 20:21'!
portNumber
	"Answer the port number for the Enchanting server."

	^ 42002   "Scratch runs on 42001"
! !
